function [retval1, retval2, retval3]=dsyjf2a(H)
% DSYJF2A: Decomposition H = G * JJ * G', where H is a non-zero
%         real symmetric matrix, JJ = diag(+-1), 
%         and G has full column rank.
%         On exit vector P contains the pivoting which took place.
% Usage: [G,JJ,P]=dsyjf2(H)
% Standard precision (double).
% Uses unequilibrated diagonal pivoting.
% For details see: 
%    I. Slapnicar: Analysis of real symmetric and Hermitian
%    decomposition, Tech. report, University of Split, 1995.
%    (ivan.slapnicar@fesb.hr)             
%    Differences to dsyjf2 are pointed out.
[n,m]=size(H);
%%%  Initialize alpha, npos and r, and the vectors J and P.   
alpha = ( 1 + sqrt(17) ) / 8;
i = 1;
npos = 0;
r = 0;
P = 1:n;
JJ=1:n;
%%%  Main loop.
while i <= n,
%%%  Make the current submatrix symmetric.
   in=i:n;
   if (n-i)>0; H(in,in) = tril(H(in,in)) + tril(H(in,in),1)'; end;
%%%  Find nu0 and nu1 and the indices p, q and s.
   [temp,p] = max(abs(H(in,in)-diag(diag(H(in,in)))));
   [nu0,q] = max(temp);
   p = p(q);
   [nu1,s] = max(abs(diag(H(in,in))));
   if nu1 >= alpha * nu0,
%%%  1 x 1 pivot. If the block currently being reduced is singular, 
%%%  the algorithm is finished.
      if nu1 == 0,
         r = i - 1;
         i = n + 1;
      else
%%%  Bring the element H(s,s) to the position (i,i), and notify the 
%%%  permutation in the vector P.
         s = s + i - 1;
         H([i s],:) = H([s i],:);
         H(in,[i s]) = H(in,[s i]);
         P([i s]) = P([s i]);
%%%  Update H.
         JJ(i) = sign(H(i,i));
         npos = npos + (1 + JJ(i)) / 2;
         H(i,i) = sqrt(abs(H(i,i)));
         if i < n,
            i1n=i+1:n;
            H(i1n,i) = H(i1n,i) * ( JJ(i) / H(i,i) );
            H(i,i1n) = zeros(1,n-i);
            H(i1n,i1n) = H(i1n,i1n) - JJ(i) * H(i1n,i) * ...
                             H(i1n,i)';
         end;
         i = i + 1;
      end;
   else
%%%  2 x 2 pivot. Bring the elements H(p,p) and H(q,q) to the
%%%  positions H(i+1,i+1) and H(i,i), respectively, and notify the
%%%  permutation in vector P.
      p = p + i - 1;
      q = q + i - 1;
      H([i q],:) = H([q i],:);
      H([i+1 p],:) = H([p i+1],:);
      H(in,[i q]) = H(in,[q i]);
      H(in,[i+1 p]) = H(in,[p i+1]);
      P([i q]) = P([q i]);
      P([i+1 p]) = P([p i+1]);
%%%  Compute the 2 x 2 orthogonal matrix.
      z = ( H(i+1,i+1) - H(i,i) ) / ( 2 * H(i+1,i) );
      signz=sign(z); if signz==0; signz=1;end;
      t = signz / ( abs(z) + sqrt( z^2 + 1 ) );
      h = sqrt( 1 + t^2 );
      cs = 1 / h;
      sn = t / h;
      Q = [cs sn; -sn cs];
%%%  FIRST DIFFERENCE TO dsyjf2
%%%  Compute 2 x 2 inverse matrix
      X = [H(i+1,i+1) -H(i+1,i);-H(i+1,i) H(i,i)] / ... 
         ( H(i+1,i+1) * H(i,i) - H(i+1,i)*H(i+1,i));
%%%  Updating H.
      a = H(i,i) - H(i+1,i) * t;
      b = H(i+1,i+1) + H(i+1,i) * t;
      JJ(i:i+1) = sign([a b]);
      npos = npos + 1;
      D = sqrt(diag(abs([a b])));
      H(i:i+1,i:i+1) = Q * D;
      if i < n-1,
         i2n=i+2:n;
         H(i:i+1,i2n) = zeros(2,n-i-1);
%%%  SECOND DIFFERENCE TO dsyjf2
         H(i2n,i2n) = H(i2n,i2n) - H(i2n,i:i+1) * ... 
                          X * H(i2n,i:i+1)';
         H(i2n,i:i+1) = H(i2n,i:i+1) * (Q * diag(JJ(i:i+1)) * inv(D));
      end;
      i = i + 2;
   end;      
end;
%%%  If non-singularity did not occur, then rank = dimension.
if r == 0,
   r = n;
end;
%%%  Permuting the columns of H to sort J.
% i = npos + 1;
% for k = 1:npos,
%   if JJ(k) == -1,
%      while JJ(i) == -1,
%         i = i + 1;
%      end;
%      H(1:n,[i k]) = H(1:n,[k i]);
%      i = i + 1;
%   end;
% end;
retval2=diag(JJ(1:r));
H(P,1:r) = H(:,1:r);
retval1=H(:,1:r);
retval3=P;










