function [D, X, iter,h] = indef_m_scheme_brzoG_21crit(G, B, T, T2, X, n_pos, n_neg, tol, maxiter, msch,opts)
% indef_m_scheme_brzoG(G, B, T, T2, X, n_pos, n_neg, tol, maxiter, opts) differs
% from indef_m_scheme_brzo(A, B, T, T2, X, n_pos, n_neg, tol, maxiter, opts) in
% A = G'*G

% Outputs
% =======  
%
%     D      array (nev-by-1)     computed n2=n_neg largest B-negative and n1=n_pos
%                                 smallest B-positive eigenvalues, 
%                                 arranged in ascending order, i.e.
%                                 D(1:n2) are B-negative, and
%                                 D(1)<=D(2)<=...<=D(n2)
%                                 D(n2+1:n2+n1) are B-positive, and
%                                 D(n2+1)<=D(n2+2)<=...<=D(n2+n1)
%
%     X      array (n-by-nev)    computed corresponding eigenvectors    
%     h   struct               iterative history   
%            h.res_neg (iter+1-by-n2)  h.res_neg(i,:) records the ith set normalized residuals (in 2-norm)
%                                       ||AX(:, j) - D0(j)*BX(:, j)||/( (normA + abs(D0(j))*normB) * ||X(:, j)|| ) 
%                                       j=1:n2; 
%                                       normA =  the 1-norm of the matrix A
%                                       normB = the 1-norm of the matrix B
%            h.res_pos (iter+1-by-n1)  h.res_pos(i,:) records the ith set normalized residuals (in 2-norm)
%                                       ||AX(:, j) - D0(j)*BX(:, j)||/( (normA + abs(D0(j))*normB) * ||X(:, j)|| ) 
%                                       j=n2+(1:n1)
%            h.eigen_neg  (iter+1-by-n2)  h.eigen_neg(i,:) records the ith set of approximate B-negative eigenvalues.
%            h.eigen_pos  (iter+1-by-n1)  h.eigen_pos(i,:) records the ith set of approximate B-positive eigenvalues.
%            h.step_neg
%            h.step_pos
%
%--------------------------------------------------------------------------------------------------------- 
[n, ncol] = size(X);
n1=n_pos;n2=n_neg;
nev=n1+n2;
indX=1:nev;
normA = norm((G')*G, 1);
normB = norm(B, 1);

PHH = [];PH=[];
APH = [];APHH=[];
BPH = [];BPHH=[];

if nargin ==10
  if strcmp(class(T), 'double'),
    % funT = @(x) T\x;
     funG= @(x) G\( (G')\ x);
     funT=funG;
  else
    % funT = @(x) feval(T, x);
    funG = @(x) feval(T, x);
    funT=funG;
  end
   funT_pos = funT;
   if size(T2) == 0,
        funT2 = @(x) [];
        funT_neg = funT;
   elseif strcmp(class(T2), 'double'),
       funT2 = @(x) T2\x;
       funT_neg = funT2;
   else
       funT2 = @(x) feval(T2, x);
       funT_neg = funT2;
   end
elseif nargin==11,
    %% CG based preconditioner
    if strcmp(class(T), 'double'),
         funT_pos = T;
    end
    if size(T2) == 0,
         funT_neg = funT_pos;
    elseif strcmp(class(T2), 'double'),
         funT_neg = T2;
    end
    cgopts.nitn = 50;
    cgopts.tol  = 1.0e-2;
    cgopts.met  = 0;
    cgx0=zeros(n,1);
    cgM=0;
    h.cgiter=[0 0];
    h.T_pos = [];
    h.T_neg = [];
end
if nargin <10
    msch='whole';
end



% Initialization
AX = G'*(G*X); BX = B*X;
[X, AX, BX, d] = MGS2G(X, AX, BX, G, B); % Bp=X'*BX=diag(d)
m1 = sum(d>0);
m2 = sum(d<0);
if (m2 < n2),
    error('Initial X does not have at least %d B-negative columns.', n2);
end
if (m1 < n1),
    error('Initial X does not have at least %d B-positive columns.', n1);
end
Ap = X'*AX;
Bp = X'*BX;
Ap = (Ap+Ap') / 2; Bp = (Bp+Bp') / 2;
[V0, D0] = eig(Ap, Bp);
D0 = diag(D0);
D0 = real(D0)+abs(imag(D0));
[D0, ind] = sort(D0);
V0 = real(V0(:, ind));
D0=D0(m2-n2+indX);
V0=V0(:,m2-n2+indX);
X = X*V0;
AX = AX*V0; BX=BX*V0;
d = 0*D0;
d(1:n2) = -1;
d(n2+1:end) = 1;

D = zeros(nev, 1);
%%
% This algorithm finds n1 smallest eigenvalues with positive sign
% and n2 largest eigenvalues with negative sign, i.e. the inertia
% of X'BX keeps unchanged.

ee_neg=zeros(1,n2);
re_neg=zeros(1,n2);
ee_pos=zeros(1,n1);
re_pos=zeros(1,n1);
temp_re=[];
temp_ev=[];
step_neg=[];
step_pos=[];
deflated = 0;
deflated_pos = 0;
deflated_neg = 0;
tr_def = 0;
dflt = 0;
if dflt == 1,
    def_thresh = 5;
else
    def_thresh = maxiter;
end

for iter = 1 : maxiter
    % Check converged eigenvalues
    normD = norm(D);
    maxr_neg = 0;%
    dflt_neg = 0;%
    j = deflated + n2 - deflated_neg;
    jj=1;
    while j > deflated                % from the middle to the left
        R(:, j) = AX(:, j) - D0(j)*BX(:, j);
        r = norm(R(:, j), 2) / (normA + abs(D0(j))*normB) / norm(X(:, j), 2);
  %      r = norm(R(:, j), 2) / norm(X(:, j), 2) / abs(D0(j)) / normB;
        if (deflated_neg==0)
        ee_neg(j) = D0(j);
        re_neg(j) = r;
        end
        if dflt,
            if r > maxr_neg, maxr_neg = r; end              %lambda_2^- cannot deflate before lambda_1^- does, etc.
        else
            if j > n2+deflated_pos-n_neg && r > maxr_neg, maxr_neg = r; end              %lambda_2^- cannot deflate before lambda_1^- does, etc.
            if maxr_neg < tol, dflt_neg = dflt_neg+1; end
        end
        % deflate ``negative'' eigenvalues
        if iter > def_thresh && maxr_neg <= tol && deflated_neg < n2,
            deflated_neg = deflated_neg+1;
            deflated = deflated+1;
            D(deflated) = D0(j);
            tr_def = tr_def - D0(j);
            fprintf('%d. neg. has converged in %d. step\n',deflated_neg,iter)
            step_neg=[step_neg,[deflated_neg;iter]];
                 ee_neg(n2+1-deflated_neg)=D0(j);
                 re_neg(n2+1-deflated_neg) = r;
            if (j > deflated),
                 X(:, [deflated, deflated+1:j]) =  X(:, [j, deflated:j-1]);
                AX(:, [deflated, deflated+1:j]) = AX(:, [j, deflated:j-1]);
                BX(:, [deflated, deflated+1:j]) = BX(:, [j, deflated:j-1]);
                 P(:, [deflated, deflated+1:j]) =  P(:, [j, deflated:j-1]);
                AP(:, [deflated, deflated+1:j]) = AP(:, [j, deflated:j-1]);
                BP(:, [deflated, deflated+1:j]) = BP(:, [j, deflated:j-1]);
                 R(:, [deflated, deflated+1:j]) =  R(:, [j, deflated:j-1]);
                   D0([deflated, deflated+1:j]) =    D0([j, deflated:j-1]);
                    d([deflated, deflated+1:j]) =     d([j, deflated:j-1]);
            end
         else
            temp_re(jj)=r;
            temp_ev(jj)=D0(j);
            jj=jj+1;
            j = j - 1;
         end
    end
    s=size(temp_ev,2);
    if s>0
          ee_neg(1:s)=temp_ev(s:-1:1);
          re_neg(1:s)=temp_re(s:-1:1);
    end
    temp_ev=[];temp_re=[];
    if iter>1
         eigen_neg=[eigen_neg;ee_neg(1:size(eigen_neg,2))];
         res_neg=[res_neg;re_neg(1:size(res_neg,2))];
    else
         eigen_neg=ee_neg;
         res_neg=re_neg;
    end
    maxr_pos = 0;
    dflt_pos = 0;
    jj=1;
    for j = deflated + n2 - deflated_neg + 1 : nev  % from the middle to the right
        R(:, j) = AX(:, j) - D0(j)*BX(:, j);
        r = norm(R(:, j), 2) / (normA + abs(D0(j))*normB) / norm(X(:, j), 2);
%        r = norm(R(:, j), 2) / norm(X(:, j), 2) / abs(D0(j)) / normB;
        if (deflated_pos==0)
            ee_pos(j-n2) = D0(j);
            re_pos(j-n2) = r;
        end
        if dflt == 1,
            if r > maxr_pos, maxr_pos = r; end%
        else
            if j <= n2+n_pos && r > maxr_pos, maxr_pos = r; end%
            if maxr_pos < tol, dflt_pos = dflt_pos+1; end
        end
        % deflate ``positive'' eigenvalues
        if iter > def_thresh && maxr_pos <= tol && deflated_pos < n1,
            deflated_pos = deflated_pos+1;
            deflated = deflated+1;
            D(deflated) = D0(j);
            tr_def = tr_def + D0(j);
            fprintf('%d. pos. has converged in %d. step\n',deflated_pos,iter)
             step_pos=[step_pos,[deflated_pos;iter]];
                   ee_pos(deflated_pos)=D0(j);
                   re_pos(deflated_pos)=r;
            if (j > deflated),
                X(:, [deflated, j]) = X(:, [j, deflated]);
                AX(:, [deflated, j]) = AX(:, [j, deflated]);
                BX(:, [deflated, j]) = BX(:, [j, deflated]);
                P(:, [deflated, j]) = P(:, [j, deflated]);
                AP(:, [deflated, j]) = AP(:, [j, deflated]);
                BP(:, [deflated, j]) = BP(:, [j, deflated]);
                R(:, [deflated, j]) = R(:, [j, deflated]);
                D0([deflated, j]) = D0([j, deflated]);
                d([deflated, j]) = d([j, deflated]);
            end
        else
            temp_re(jj)=r;
            temp_ev(jj)=D0(j);
            jj=jj+1;
        end
    end
    d(1:deflated) = 0;
    s=size(temp_ev,2); % # of B-neg. ev's which did not converge in this iteration
    if s>0
          ee_pos(deflated_pos+1:end)=temp_ev(1:s);
          re_pos(deflated_pos+1:end)=temp_re(1:s);
    end
    temp_re=[];temp_ev=[];
    if iter>1
          eigen_pos=[eigen_pos;ee_pos];
          res_pos=[res_pos;re_pos];
    else
          eigen_pos=ee_pos;
          res_pos=re_pos;
    end
    if dflt == 1, dflt_neg = deflated_neg; dflt_pos = deflated_pos; end
     if iter==1,
       fprintf('step %3d,  deflated %2d(%2d+%2d),  r_neg = %.2e, r_pos = %.2e\n',...
           iter, deflated, dflt_neg, dflt_pos, maxr_neg, maxr_pos);
     else
         fprintf('step %3d,  deflated %2d(%2d+%2d),  r_neg = %.2e, r_pos = %.2e, trace=%.10e\n',...
           iter, deflated, dflt_neg, dflt_pos, maxr_neg, maxr_pos,tr);
     end
    if deflated >= nev, break; end
    if max(maxr_neg, maxr_pos) <= tol, break; end
    indX = 1 : nev-deflated;
    active = deflated + indX;
    act_pos = find(d>0)';
    act_neg = find(d<0)';

    % Generate search subspace
    if nargin == 10,
%       W = [funT(R(:, active)), funT2(R(:, active))];
        W = [funT_pos(R(:, act_pos)), funT_neg(R(:, act_neg))];
    elseif nargin==11,
        W = [];
%        for i = active,
%            [W0, cg_error, cg_iter, cg_flag] = LinCG(T, cgx0, R(:, i), cgM, cgopts);
%            h.K_CGITER=[h.K_CGITER,cg_iter];
%            W = [W, W0];
%            h.cgiter(1)=h.cgiter(1)+cg_iter;
%            if size(T2)>0,
%            [W0, cg_error, cg_iter, cg_flag] = LinCG(T2, cgx0, R(:, i), cgM, cgopts);
%            h.M_CGITER=[h.M_CGITER,cg_iter];
%            W = [W, W0];
%            h.cgiter(2)=h.cgiter(2)+cg_iter;
%            end
%        end
        for i = act_pos,
            [W0, cg_error, cg_iter, cg_flag] = LinCG(funT_pos, cgx0, R(:, i), cgM, cgopts);
            h.T_pos=[h.T_pos,cg_iter];
            W = [W, W0];
            h.cgiter(1)=h.cgiter(1)+cg_iter;
        end
        for i = act_neg,
            [W0, cg_error, cg_iter, cg_flag] = LinCG(funT_neg, cgx0, R(:, i), cgM, cgopts);
            h.T_neg=[h.T_neg,cg_iter];
            W = [W, W0];
            h.cgiter(2)=h.cgiter(2)+cg_iter;
        end
    end
    AW = G'*(G*W); BW = B*W;
   %% Search subspace
        if iter == 1 || (isnumeric(msch) && msch==2),
             V = [X, W];
             AV = [AX, AW]; BV = [BX, BW];
        else
          if isnumeric(msch)
                 % msch=2,3,4,5,...
% %                  if msch==2 % PSD/A or BPSD/A
% %                       PHH=[];APHH=[];BPHH=[];
                 %else % using previous iterates in search subspace
             if msch==3
                 PHH=P(:, active);APHH = AP(:,active);BPHH = BP(:,active); 
             else % msch>3 
                PH = [P(:, active) PHH];          
                if  iter<=(msch-2)
                    % not enough previous directions
                    PHH=PH;
                    APH = [AP(:, active) APH];
                    APHH=APH; 
                    BPH = [BP(:, active) BPH];
                    BPHH=BPH;
                else
                    % e.g. msch=4;iter=3; PH=[P^(3),P^(2),P^(1)]
                    % e.g. msch=4;iter=3; PHH=[P^(3),P^(2)]
                    sizeact=size(active,2);
                    PHH=PH(:,1:(msch-2)*sizeact); 
                    APH = [AP(:, active) APH];
                    APHH=APH(:,1:(msch-2)*sizeact); 
                    BPH = [BP(:, active) BPH];
                    BPHH=BPH(:,1:(msch-2)*sizeact);
               end %iter<=(msch-2)
            end % msch==3  
          else % msch is some string
               % whole history: all previous directions
               sizeact=size(active,2);
               PHH = [P(:, active) PHH];
               APH = [AP(:, active) APH];
               APHH=APH;
               BPH = [BP(:, active) BPH];
               BPHH=BPH;%(:,1:(msch-2)*sizeact);
         end % isnumeric(msch)
         V=[X,PHH,W]; % whole X (first columns are converged columns if deflated>0)
         AV=[AX,APHH,AW]; BV=[BX,BPHH,BW];
        end
% % %        perm = [];
% % %        iperm(perm) = 1:size(V, 2);
% % %        AV = A*V; BV = B*V;
% %         m = 2*(nev-deflated);
% % %        indP = nev + active;
% %     else
% %         V = [X, P(:, active), W];                 % whole X (first columns are converged columns if deflated>0)
% %         AV = [AX, AP(:, active), AW];
% %         BV = [BX, BP(:, active), BW];
% % %        AV = A*V; BV = B*V;
% %         m = 3*(nev-deflated);
% % %        indP = [nev + active, 2*nev-deflated + active];
% %     end
%%
     [V, AV, BV, d, m] = MGS2G(V, AV, BV, G, B, size(X, 2));
    while m <= n1+n2,
        V = [X, randn(size([P(:, active), W]))];
        [V, AV, BV, d, m] = MGS2G(V, AV, BV, G, B, size(X, 2));
    end
    m1 = sum(d>0);
    m2 = sum(d<0);
%    fprintf('m1=%d, m2=%d\n',m1,m2)
    if (m1 < n1 || m2 < n2),  
        error('This will never happen in exact arithmetic.');
    end

    % Exclude converged eigenvectors
    V = V(:, deflated+1 : end);            % X is without converged columns
    AV = AV(:, deflated+1 : end);
    BV = BV(:, deflated+1 : end);
    % Use explicit matrix multiplication every 10 steps
    if (mod(iter, 10) == 0),
        AV = G'*(G * V);
        BV = B * V;
    end
    % Solve projected problem
    Ap = V'*AV;
    Bp = V'*BV; % In fact, Bp = diag{\pm 1};
%    Bp
    Ap = (Ap+Ap') / 2; Bp = (Bp+Bp') / 2;
    [V0, D0] = eig(Ap, Bp);
    D0 = diag(D0);
    if norm(imag(D0))>0,
        warning('Complex eigenvalues detected!');
    end
    D0 = real(D0)+imag(D0);
    % Extract n1-deflated_pos smallest ``positive'' eigenvalues
    % and n2-deflated_neg largest ``negative'' eigenvalues
    [D0, ind] = sort(D0);
    V0 = real(V0(:, ind));     % so neg.columns are not mixed in deflation of B-pos. ev's:              
                                             % if (j > deflated),
                                             % X(:, [deflated, j]) = X(:, [j, deflated]);
    m1 = deflated_pos; m2 = deflated_neg;
    for j = 1:size(Bp),
        tmp = V0(:, j)'*Bp*V0(:, j);
        if tmp > 0,
            m1 = m1+1;
        elseif tmp < 0,
            m2 = m2+1;
        end
    end
    D0(active) = D0(m2-n2 + indX);
    d = zeros(nev, 1);
    d(deflated+(1:n2-deflated_neg)) = -1;
    d(deflated+1+n2-deflated_neg:end) = 1;
    DD=D0(active);
    tr_neg=sum(DD(1:n2-deflated_neg));
    tr_pos=sum(DD(n2-deflated_neg+1:end));
    tr=tr_pos-tr_neg+tr_def;
    trace=[tr_neg tr_pos tr tr_def];
    V0 = V0(:, m2-n2 + indX);
    indP = nev-deflated+1 : m-deflated;

    % Update P and X
    P(:, active) = V(:, indP) * V0(indP, :);
    AP(:, active) = AV(:, indP) * V0(indP, :);
    BP(:, active) = BV(:, indP) * V0(indP, :);
    X(:, active) = P(:, active) + V(:, indX) * V0(indX, :);
    AX(:, active) = AP(:, active) + AV(:, indX) * V0(indX, :);
    BX(:, active) = BP(:, active) + BV(:, indX) * V0(indX, :);
end

if deflated < nev,       % not all converged in maxiter iterations
    D(deflated+1:end)=D0(active);
end
[D, ind] = sort(D);
X = X(:, ind);
% last rel. residuals and eigenvalues
for j = n2:-1:1                               % from the middle to the left
        R(:, j) = G'*(G*X(:, j)) - D(j)*B*X(:, j);
        r = norm(R(:, j), 2) / (normA + abs(D(j))*normB) / norm(X(:, j), 2);
%        r = norm(R(:, j), 2) / norm(X(:, j), 2) / abs(D0(j)) / normB;
        ee_neg(j) = D(j);
        re_neg(j) = r;
end
h.eigen_neg=[eigen_neg;ee_neg(1:size(eigen_neg,2))];
h.res_neg=[res_neg;re_neg(1:size(res_neg,2))];
for j = n2+(1:n1)                            % from the middle to the right
       R(:, j) = G'*(G*X(:, j)) - D(j)*B*X(:, j);
       r = norm(R(:, j), 2) / (normA + abs(D(j))*normB) / norm(X(:, j), 2);
 %      r = norm(R(:, j), 2) / norm(X(:, j), 2) / abs(D0(j)) / normB;
       ee_pos(j-n2) = D(j);
       re_pos(j-n2) = r;
end
 h.eigen_pos=[eigen_pos;ee_pos];
 h.res_pos=[res_pos;re_pos];
 h.step_neg=step_neg;
 h.step_pos=step_pos;
 if nargin==10
     h.T_pos=[];h.T_neg=[];
 end
return
